import contextlib
import os
import sys
from functools import wraps
from multiprocessing import cpu_count
from subprocess import check_call

from rich.console import Console

# setup a very large console to prevent rich from wrapping long error messages
# since it usually has the side effect of truncating file names.
# This workaround also helps keeping error messages reproducible in CI.


def print_err(message: str) -> None:
    err_console = Console(width=500, file=sys.stderr)
    err_console.print(f"[bold white on red]ERROR[/] {message}")


def requires_idefix(func):
    @wraps(func)
    def _func(*args, **kwargs):
        if os.getenv("IDEFIX_DIR") is None:
            print_err(
                "this functionality requires $IDEFIX_DIR to be defined.",
            )
            exit(1)
        return func(*args, **kwargs)

    return _func


@contextlib.contextmanager
def pushd(new_dir):
    previous_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(previous_dir)


def _make(directory):
    ncpus = str(min(8, cpu_count()))
    with pushd(directory):
        check_call(["make", "-j", ncpus])



from pathlib import Path
from subprocess import check_call
from uuid import uuid4

from inifix.idefix_conf import IdefixConf

from idefix_cli._commons import _make, print_err, pushd


def _add_run_args(parser):

    parser.add_argument("directory", help="target directory")
    parser.add_argument(
        "-i",
        dest="inifile",
        action="store",
        type=str,
        default="idefix.ini",
        help="target inifile",
    )
    time_group = parser.add_mutually_exclusive_group()
    time_group.add_argument(
        "--duration",
        action="store",
        type=float,
        help="run for specified time (in code units)",
    )
    time_group.add_argument(
        "--one",
        "--one-step",
        dest="one_step",
        action="store_true",
        help="run only for one time step",
    )
    parser.add_argument(
        "--time-step",
        dest="time_step",
        action="store",
        type=float,
        help="patch the inifile TimeIntegrator.first_dt parameter",
    )


def _get_patched_inifile(
    directory: str,
    inifile: str = "idefix.ini",
    duration: float = None,
    time_step: float = None,
) -> Path:
    conf = IdefixConf(inifile)

    # conf["Output"]["dmp"] = 0.0  # force dump file to be written at every timestep

    to_write = False
    if time_step not in (None, conf["TimeIntegrator"]["first_dt"]):
        conf["TimeIntegrator"]["first_dt"] = time_step
        to_write = True
    if duration not in (None, conf["TimeIntegrator"]["tstop"]):
        conf["TimeIntegrator"]["tstop"] = duration
        to_write = True

    if not to_write:
        return Path(inifile)
    name = "{}_{}.ini_".format(inifile.stem, uuid4())
    new_inifile = Path(directory) / name

    conf.write(new_inifile)
    return new_inifile


def run(
    directory: str,
    inifile: str = "idefix.ini",
    duration: float = None,
    time_step: float = None,
    one_step: bool = False,
) -> int:

    input_inifile = inifile
    for loc in [Path.cwd(), Path(directory)]:
        inifile = (loc / input_inifile).resolve()
        if inifile.is_file():
            break
    else:
        print_err("could not find inifile {input_inifile}")
        return 1
    if one_step:
        if time_step is None:
            time_step = IdefixConf(inifile)["TimeIntegrator"]["first_dt"]
        duration = time_step

    d = Path(directory)
    if not (d / "idefix").is_file():
        if not (d / "Makefile").is_file():
            print_err(
                "No idefix instance or Makefile found in the target directory. "
                "Run `idfx setup` first."
            )
            return 1
        _make(directory)

    if (duration, time_step) != (None, None):
        ninifile = _get_patched_inifile(directory, inifile, duration, time_step)
        if ninifile != inifile:
            inifile = ninifile
            print("Running patched inifile {}".format(inifile))

    with pushd(d):
        check_call(["./idefix", "-i", inifile])
    return 0



import argparse
import os
import re
from pathlib import Path

from idefix_cli._commons import _make, print_err, pushd, requires_idefix

cpuarch = frozenset(("HSW", "BDW", "SKX", "EPYC"))

gpuarch = frozenset(
    (
        "Kepler30",
        "Maxwell50",
        "Pascal60",
        "Pascal61",
        "Volta70",
        "Volta72",
        "Turing75",
    )
)
known_archs = {"CPU": cpuarch, "GPU": gpuarch}
default_cpu_arch = "BDW"
default_gpu_arch = "Pascal60"
SETUP_DEFAULTS = {"arch": [default_cpu_arch, default_gpu_arch]}


def _add_setup_args(parser):
    parser.add_argument("directory", help="target directory")

    parser.add_argument("-mhd", action="store_true", help="enable MHD")
    parser.add_argument(
        "-gpu", dest="use_gpu", action="store_true", help="enable KOKKOS+CUDA"
    )

    parser.add_argument("-cxx", help="override default compiler")

    parser.add_argument(
        "-arch",
        nargs="+",
        default=SETUP_DEFAULTS["arch"],
        choices=cpuarch.union(gpuarch),
        help="Target Kokkos architecture",
    )
    parser.add_argument(
        "-openmp", help="enable OpenMP parallelism", action="store_true"
    )
    parser.add_argument("-mpi", help="enable MPI parallelism", action="store_true")
    parser.add_argument(
        "--make",
        "--compile",
        action="store_true",
        help="also compile idefix executable",
    )


@requires_idefix
def setup(
    directory: str,
    arch: str = SETUP_DEFAULTS["arch"],
    use_gpu: bool = False,
    cxx: str = None,
    openmp: bool = False,
    mpi: bool = False,
    mhd: bool = False,
    make: bool = False,
):
    """This is the $IDEFIX_DIR/configure.py script in a reusable form"""

    makefileOptions = {
        "extraIncludeDir": "",
        "extraVpath": "",
        "extraHeaders": "",
        "extraObj": "",
        "extraLine": "",
        "cxxflags": "",
        "ldflags": "",
    }

    # extract cpu & gpu architectures from arch
    selected_archs = {"CPU": default_cpu_arch, "GPU": default_gpu_arch}

    for core_type, archs in known_archs.items():
        vals = list(archs.intersection(set(arch)))
        if not vals:
            continue
        if len(vals) > 1:
            print_err(f"received more than one {core_type} arch: {vals} ")
            return 1
        selected_archs[core_type] = vals[0]

    # todo: remove those confusing aliases and rename `use_gpu` into `gpu`
    cpu = selected_archs["CPU"]
    gpu = selected_archs["GPU"]

    if use_gpu:
        makefileOptions.update(
            {
                "extraLine": '\nKOKKOS_CUDA_OPTIONS = "enable_lambda"',
                "cxx": "${KOKKOS_PATH}/bin/nvcc_wrapper",
                "kokkosDevices": '"Cuda"',
                "kokkosArch": cpu + "," + gpu,
                "cxxflags": "-O3 ",
            }
        )

        # Enforce backend compiler for nvcc
        add = "\nexport NVCC_WRAPPER_DEFAULT_COMPILER = "
        if cxx:
            makefileOptions["extraLine"] += add + cxx
        elif mpi:
            makefileOptions["extraLine"] += add + "mpicxx"
    else:
        if cxx:
            makefileOptions["cxx"] = cxx
        else:
            if mpi:
                makefileOptions["cxx"] = "mpicxx"
            else:
                makefileOptions["cxx"] = "g++"

        makefileOptions["kokkosArch"] = cpu
        makefileOptions["cxxflags"] = "-O3"
        if openmp:
            makefileOptions["kokkosDevices"] = '"OpenMP"'
        else:
            makefileOptions["kokkosDevices"] = '"Serial"'

    if mpi:
        makefileOptions["extraIncludeDir"] += " -I$(SRC)/dataBlock/mpi"
        makefileOptions["extraVpath"] += ":$(SRC)/dataBlock/mpi"
        makefileOptions["extraHeaders"] += " mpi.hpp"
        makefileOptions["extraObj"] += " mpi.o"
        makefileOptions["cxxflags"] += " -DWITH_MPI"

    if mhd:
        makefileOptions["extraIncludeDir"] += " -I$(SRC)/hydro/MHDsolvers"
        makefileOptions["extraVpath"] += ":$(SRC)/hydro/MHDsolvers"
        makefileOptions["extraHeaders"] += " solversMHD.hpp"
        makefileOptions["cxxflags"] += " -DMHD=YES"
    else:
        makefileOptions["extraIncludeDir"] += " -I$(SRC)/hydro/HDsolvers"
        makefileOptions["extraVpath"] += ":$(SRC)/hydro/HDsolvers"
        makefileOptions["extraHeaders"] += " solversHD.hpp"
        makefileOptions["cxxflags"] += " -DMHD=NO"

    # Makefile substitution
    data = (Path(os.getenv("IDEFIX_DIR")) / "Makefile.in").read_text()

    for key, val in makefileOptions.items():
        data = re.sub(r"@{0}@".format(key), val, data)

    with pushd(directory):
        Path("Makefile").write_text(data)

    # print information
    def _print_status(name: str, flag: bool):
        prefix = {True: "en", False: "dis"}[flag]
        print("{}: {}abled".format(name, prefix))

    print("-----------------------------------------------------------")
    print("Idefix succesfully configured with the following options:\n")
    _print_status("MHD", mhd)

    print("Compiler: {}".format(makefileOptions["cxx"]))
    _print_status("MPI", mpi)
    if use_gpu:
        print("Execution target: GPU")
        print("Target architecture: " + gpu)
    else:
        print("Execution target: CPU")
        print("Target architecture: " + cpu)
        _print_status("OpenMP", openmp)

    print("Cflags: {}".format(makefileOptions["cxxflags"]))
    print("-----------------------------------------------------------")
    if make:
        _make(directory)
    return 0


def main():
    parser = argparse.ArgumentParser("Setup an Idefix problem.")
    _add_setup_args(parser)
    args = parser.parse_args()

    return setup(
        args.directory,
        args.arch,
        args.use_gpu,
        args.cxx,
        args.openmp,
        args.mpi,
        args.mhd,
        args.make,
    )


if __name__ == "__main__":  # pragma: no coverage
    exit(main())



import argparse
import sys
from typing import Optional

from idefix_cli import __version__
from idefix_cli.commands._clean import _add_clean_args, clean
from idefix_cli.commands._read import _add_read_args, read
from idefix_cli.commands._run import _add_run_args, run
from idefix_cli.commands._setup import _add_setup_args, setup
from idefix_cli.commands._stamp import _add_stamp_args, stamp
from idefix_cli.commands._write import _add_write_args, write


def main(argv: Optional[list[str]] = None) -> int:
    parser = argparse.ArgumentParser(prog="idfx")
    parser.add_argument("-v", "--version", action="version", version=__version__)

    subparsers = parser.add_subparsers(dest="command")

    clean_parser = subparsers.add_parser("clean", help="clean up generated files")
    _add_clean_args(clean_parser)

    read_parser = subparsers.add_parser(
        "read", help="read an Idefix inifile and print it to json format"
    )
    _add_read_args(read_parser)

    run_parser = subparsers.add_parser("run", help="run an Idefix problem")
    _add_run_args(run_parser)

    setup_parser = subparsers.add_parser("setup", help="setup an Idefix problem")
    _add_setup_args(setup_parser)

    stamp_parser = subparsers.add_parser(
        "stamp", help="print relevant data for reproduction to stdout"
    )
    _add_stamp_args(stamp_parser)

    write_parser = subparsers.add_parser(
        "write", help="write an Idefix inifile a from json string"
    )
    _add_write_args(write_parser)

    args = parser.parse_args(argv)

    if args.command is None:
        # calling `idfx` without any argument is equivalent to `idfx --help`
        # except that the return value non zero.
        parser.print_help(sys.stderr)
        return 1

    if args.command == "clean":
        return clean(args.directory, args.all, args.dry)

    if args.command == "read":
        return read(args.inifile, args.indent)

    if args.command == "run":
        return run(
            args.directory,
            args.inifile,
            args.duration,
            args.time_step,
            args.one_step,
        )

    if args.command == "setup":
        return setup(
            args.directory,
            args.arch,
            args.use_gpu,
            args.cxx,
            args.openmp,
            args.mpi,
            args.mhd,
            args.make,
        )

    if args.command == "stamp":
        return stamp(args.todict)

    if args.command == "write":
        return write(args.dest, args.source, args.force)



import json
import os
from datetime import datetime
from getpass import getuser
from socket import gethostname
from time import ctime

import git

from idefix_cli._commons import requires_idefix


def _add_stamp_args(parser):
    parser.add_argument(
        "-d",
        "--json",
        dest="todict",
        action="store_true",
        help="print to JSON serializable format.",
    )


@requires_idefix
def stamp(todict: bool = False):
    """
    Print idefix latest version tag-(git hash) and current time to stdout.
    """
    repo = git.Repo(os.environ["IDEFIX_DIR"])
    data = {
        "tag": str(repo.tags[-1]),
        "sha": repo.head.object.hexsha,
        "user": getuser(),
        "host": gethostname(),
        "date": ctime(datetime.now().timestamp()),
    }

    if todict:
        text = json.dumps(data, indent=2)
    else:
        text = "\n".join(list(data.values()))

    print(text)
    return 0



import os
from glob import glob
from itertools import chain
from pathlib import Path

from idefix_cli._commons import pushd

# bpatterns are those targeted by `make clean`, which is equivalent to
# rm -f *.o *.cuda *.host
bpatterns = frozenset(("*.o", "*.cuda", "*.host"))

# additional generated files
kokkos_files = frozenset(
    (
        "KokkosCore_config.h",
        "KokkosCore_config.tmp",
        "KokkosCore_Config*.tmp",
        "KokkosCore_Config*.hpp",
        "libkokkos.a",
    )
)

# only cleared if `--all` flag is passed
gpatterns = frozenset(("*.ini_", "Makefile", "idefix"))


def _add_clean_args(parser):
    parser.add_argument("directory", help="the root directory to clean")
    parser.add_argument(
        "--all",
        action="store_true",
        help="also clean generated Makefiles, idefix excutable files and .ini_ files",
    )
    parser.add_argument(
        "--dry-run",
        "--dry",
        dest="dry",
        action="store_true",
        help="print a list of targets, no file is removed",
    )


def clean(directory, clean_all: bool = False, dry: bool = False):
    with pushd(directory):
        patterns = bpatterns.union(kokkos_files)
        if clean_all:
            patterns = patterns.union(gpatterns)

        targets = sorted(
            chain.from_iterable(glob(str(Path.cwd() / p)) for p in patterns)
        )

        if dry:
            if not targets:
                print("Nothing to remove.")
                return 0
            print("The following files would be removed.")
            print("\n".join(targets))
            return 0

        for t in targets:
            os.remove(t)

    return 0
